@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2015 Red Hat, Inc.
+# Copyright (C) 2023 Red Hat, Inc.
 # All rights reserved.
 #
 # License: GPL (version 3 or any later version).
@@ -16,14 +16,15 @@ import socket
 import time
 import shutil
 import logging
-# from nss import nss
 import subprocess
-from datetime import datetime, timedelta, date
-from subprocess import check_output
-from lib389.passwd import password_generate
-from lib389.lint import DSCERTLE0001, DSCERTLE0002
-from lib389.utils import ensure_str, format_cmd_list
 import uuid
+from datetime import datetime, timedelta
+from subprocess import check_output, run, PIPE
+from lib389.passwd import password_generate
+from lib389._mapped_object_lint import DSLint
+from lib389.lint import DSCERTLE0001, DSCERTLE0002
+from lib389.utils import ensure_str, format_cmd_list, DSVersion, cert_is_ca
+
 
 KEYBITS = 4096
 CA_NAME = 'Self-Signed-CA'
@@ -42,7 +43,7 @@ VALID_MIN = 61  # Days
 log = logging.getLogger(__name__)
 
 
-class NssSsl(object):
+class NssSsl(DSLint):
     def __init__(self, dirsrv=None, dbpassword=None, dbpath=None):
         self.dirsrv = dirsrv
         self._certdb = dbpath
@@ -56,18 +57,25 @@ class NssSsl(object):
         else:
             self.dbpassword = dbpassword
 
-        self.db_files = {"dbm_backend": ["%s/%s" % (self._certdb, f) for f in ("key3.db", "cert8.db", "secmod.db")],
-                         "sql_backend": ["%s/%s" % (self._certdb, f) for f in ("key4.db", "cert9.db", "pkcs11.txt")],
-                         "support": ["%s/%s" % (self._certdb, f) for f in ("noise.txt", PIN_TXT, PWD_TXT)]}
-        self._lint_functions = [self._lint_certificate_expiration,]
+        self.db_files = {group: [f"{self._certdb}/{f}" for f in files]
+                         for group, files in {"dbm_backend": ("key3.db", "cert8.db", "secmod.db"),
+                                              "sql_backend": ("key4.db", "cert9.db", "pkcs11.txt"),
+                                              "support": ("noise.txt", PIN_TXT, PWD_TXT)}.items()}
 
-    def lint(self):
-        results = []
-        for fn in self._lint_functions:
-            for result in fn():
-                if result is not None:
-                    results.append(result)
-        return results
+    @classmethod
+    def lint_uid(cls):
+        return 'tls'
+
+    def _assert_not_chain(self, pemfile):
+        # To work this out, we open the file and count how many
+        # begin key and begin cert lines there are. Any more than 1 is bad.
+        count = 0
+        with open(pemfile, 'r') as f:
+            for line in f:
+                if line.startswith('-----BEGIN PRIVATE KEY-----') or line.startswith('-----BEGIN CERTIFICATE-----'):
+                    count = count + 1
+        if count > 1:
+            raise ValueError(f"The file {pemfile} may be a chain file. This is not supported. Break out each certificate and key into unique files, and import them individually.")
 
     def _lint_certificate_expiration(self):
         """Check all the certificates in the db if they will expire within 30 days
@@ -85,11 +93,13 @@ class NssSsl(object):
                 # Expired
                 report = copy.deepcopy(DSCERTLE0002)
                 report['detail'] = report['detail'].replace('CERT', cert[0])
+                report['check'] = f'tls:certificate_expiration'
                 yield report
             elif diff_date < timedelta(days=30):
                 # Expiring within 30 days
                 report = copy.deepcopy(DSCERTLE0001)
                 report['detail'] = report['detail'].replace('CERT', cert[0])
+                report['check'] = f'tls:certificate_expiration'
                 yield report
 
     def detect_alt_names(self, alt_names=[]):
@@ -115,7 +125,6 @@ class NssSsl(object):
         :type alt_names: [str, ]
         :returns: String of the subject DN.
         """
-
         if self.dirsrv and len(alt_names) > 0:
             return SELF_ISSUER.format(GIVENNAME=self.dirsrv.get_uuid(), HOSTNAME=alt_names[0])
         elif len(alt_names) > 0:
@@ -186,10 +195,13 @@ only.
 
         # Init the db.
         # 48886; This needs to be sql format ...
-        cmd = ['/usr/bin/certutil', '-N', '-d', self._certdb, '-f', '%s/%s' % (self._certdb, PWD_TXT)]
+        cmd = ['/usr/bin/certutil', '-N', '-d', self._certdb, '-f', '%s/%s' % (self._certdb, PWD_TXT),  '-@', '%s/%s' % (self._certdb, PWD_TXT)]
         self._generate_noise('%s/noise.txt' % self._certdb)
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         self.log.debug("nss output: %s", result)
         return True
 
@@ -218,6 +230,30 @@ only.
         assert not self._db_exists()
         return True
 
+    def openssl_rehash(self, certdir):
+        """
+        Compatibly run c_rehash (on old openssl versions) or openssl rehash (on
+        new ones). Prefers openssl rehash, because openssl on versions where
+        the rehash command doesn't exist, also doesn't correctly set the return
+        code. Instead, we parse the output of `openssl version` and try to
+        figure out if we have a new enough version to unconditionally run rehash.
+        """
+        try:
+            openssl_version = check_output(['/usr/bin/openssl', 'version']).decode('utf-8').strip()
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+        rehash_available = DSVersion(openssl_version.split(' ')[1]) >= DSVersion('1.1.0')
+
+        if rehash_available:
+            cmd = ['/usr/bin/openssl', 'rehash', certdir]
+        else:
+            cmd = ['/usr/bin/c_rehash', certdir]
+        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
     def create_rsa_ca(self, months=VALID):
         """
         Create a self signed CA.
@@ -242,6 +278,7 @@ only.
             'CT,,',
             '-v',
             '%s' % months,
+            '-2',
             '--keyUsage',
             'certSigning',
             '-d',
@@ -251,8 +288,9 @@ only.
             '-f',
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
+        cmd_input = b'y\n\n'  # responses to certutil questions
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        result = ensure_str(run(cmd, check=True, stderr=PIPE, stdout=PIPE, input=cmd_input).stdout)
         self.log.debug("nss output: %s", result)
         # Now extract the CAcert to a well know place.
         # This allows us to point the cacert dir here and it "just works"
@@ -266,12 +304,13 @@ only.
             '-a',
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        certdetails = check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            certdetails = check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         with open('%s/ca.crt' % self._certdb, 'w') as f:
             f.write(ensure_str(certdetails))
-        cmd = ['/usr/bin/c_rehash', self._certdb]
-        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        self.openssl_rehash(self._certdb)
         return True
 
     def rsa_ca_needs_renew(self):
@@ -288,7 +327,10 @@ only.
             self._certdb,
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        certdetails = check_output(cmd, stderr=subprocess.STDOUT, encoding='utf-8')
+        try:
+            certdetails = check_output(cmd, stderr=subprocess.STDOUT, encoding='utf-8')
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         end_date_str = certdetails.split("Not After : ")[1].split("\n")[0]
         date_format = '%a %b %d %H:%M:%S %Y'
         end_date = datetime.strptime(end_date_str, date_format)
@@ -327,7 +369,10 @@ only.
             '-o', csr_path,
             ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
         # Sign the CSR with our old CA
         cmd = [
@@ -349,11 +394,12 @@ only.
             '%s' % months,
             ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
-        cmd = ['/usr/bin/c_rehash', self._certdb]
-        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        self.openssl_rehash(self._certdb)
 
         # Import the new CA to our DB instead of the old CA
         cmd = [
@@ -367,7 +413,10 @@ only.
             '-f', '%s/%s' % (self._certdb, PWD_TXT),
             ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
         return crt_path
 
@@ -380,7 +429,10 @@ only.
             '-f',
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
         # We can skip the first few lines. They are junk
         # IE ['',
@@ -401,6 +453,126 @@ only.
             cert_values.append(re.match(r'^(.+[^\s])[\s]+([^\s]+)$', line.rstrip()).groups())
         return cert_values
 
+    def _openssl_get_csr_subject(self, csr_dir, csr_name):
+        cmd = [
+            '/usr/bin/openssl',
+            'req',
+            '-subject',
+            '-noout',
+            '-in',
+            '%s/%s'% (csr_dir, csr_name),
+        ]
+        self.log.debug("cmd: %s", format_cmd_list(cmd))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        # Parse the subject string from openssl output
+        result = result.replace("subject=", "")
+        result = result.replace(" ", "").strip()
+        result = result.split(',')
+        result = result[slice(None, None, -1)]
+        result = ','.join([str(elem) for elem in result])
+        return result
+
+    def _openssl_get_csr_sub_alt_names(self, csr_dir, csr_name):
+        cmd = [
+            '/usr/bin/openssl',
+            'req',
+            '-noout',
+            '-text',
+            '-in',
+            '%s/%s'% (csr_dir, csr_name),
+        ]
+        self.log.debug("cmd: %s", format_cmd_list(cmd))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        subaltnames = []
+        lines = result.split('\n')
+        for line in lines:
+            if 'DNS:' in line:
+                names = line.split(',')
+                for altname in names:
+                    altname = altname.replace("DNS:", "")
+                    subaltnames.append(altname.replace(",:", "").strip())
+
+        return subaltnames
+
+    def _csr_show(self, name):
+        csr_dir = self.dirsrv.get_cert_dir()
+        result = ""
+        # Display PEM contents of a CSR file
+        if name and csr_dir:
+            if os.path.exists(csr_dir + "/" + name + ".csr"):
+                cmd = [
+                    "/usr/bin/sed",
+                    "-n",
+                    '/BEGIN NEW/,/END NEW/p',
+                    csr_dir + "/" + name + ".csr"
+                ]
+                self.log.debug("cmd: %s", format_cmd_list(cmd))
+                try:
+                    result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+                except subprocess.CalledProcessError as e:
+                    raise ValueError(e.output.decode('utf-8').rstrip())
+
+        return result
+
+    def _csr_list(self, csr_dir=None):
+        csr_list = []
+        csr_dir = self.dirsrv.get_cert_dir()
+        # Search for .csr file extensions in instance config dir
+        cmd = [
+            '/usr/bin/find',
+            csr_dir,
+            '-type',
+            'f',
+            '-name',
+            '*.csr',
+            '-printf',
+            '%f\\n',
+        ]
+        self.log.debug("cmd: %s", format_cmd_list(cmd))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        # Bail out if we cant find any .csr files
+        if len(result) == 0:
+            return []
+
+        # For each .csr file, get last modified time and subject DN
+        for csr_file in result.splitlines():
+            csr = []
+            # Get last modified time stamp
+            cmd = [
+                '/usr/bin/date',
+                '-r',
+                '%s/%s'% (csr_dir, csr_file),
+                '+%Y-%m-%d %H:%M:%S',
+            ]
+            try:
+                result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+            except subprocess.CalledProcessError as e:
+                raise ValueError(e.output.decode('utf-8').rstrip())
+
+            # Add csr modified timestamp
+            csr.append(result.strip())
+            # Use openssl to get the csr subject DN
+            csr.append(self._openssl_get_csr_subject(csr_dir, csr_file))
+            # Use openssl to get the csr subject alt host names
+            csr.append(self._openssl_get_csr_sub_alt_names(csr_dir, csr_file))
+            # Add csr name, without extension
+            csr.append(csr_file.rsplit('.', 1)[0])
+            csr_list.append(csr)
+
+        return csr_list
+
     def _rsa_cert_key_exists(self, cert_tuple):
         name = cert_tuple[0]
         cmd = [
@@ -412,11 +584,13 @@ only.
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
-
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         lines = result.split('\n')[1:-1]
         for line in lines:
-            m = re.match('\<(?P<id>.*)\> (?P<type>\w+)\s+(?P<hash>\w+).*:(?P<name>.+)', line)
+            m = re.match(r'\<(?P<id>.*)\> (?P<type>\w+)\s+(?P<hash>\w+).*:(?P<name>.+)', line)
             if name == m.group('name'):
                 return True
         return False
@@ -518,22 +692,29 @@ only.
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         self.log.debug("nss output: %s", result)
         return True
 
-    def create_rsa_key_and_csr(self, alt_names=[], subject=None):
+    def create_rsa_key_and_csr(self, alt_names=[], subject=None, name=None):
         """Create a new RSA key and the certificate signing request. This
-        request can be submitted to a CA for signing. The returned certifcate
+        request can be submitted to a CA for signing. The returned certificate
         can be added with import_rsa_crt.
         """
-        csr_path = os.path.join(self._certdb, '%s.csr' % CERT_NAME)
+        if name is None:
+            csr_path = os.path.join(self._certdb, '%s.csr' % CERT_NAME)
+        else:
+            csr_path = os.path.join(self._certdb, '%s.csr' % name)
 
         if len(alt_names) == 0:
             alt_names = self.detect_alt_names(alt_names)
         if subject is None:
             subject = self.generate_cert_subject(alt_names)
 
+        self.log.debug(f"CSR name -> {name}")
         self.log.debug(f"CSR subject -> {subject}")
         self.log.debug(f"CSR alt_names -> {alt_names}")
 
@@ -568,7 +749,10 @@ only.
             '-o', csr_path,
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
         return csr_path
 
@@ -594,7 +778,10 @@ only.
             '-c', CA_NAME,
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
         return (ca_path, crt_path)
 
@@ -608,10 +795,12 @@ only.
         assert ca is not None or crt is not None, "At least one parameter should be specified (ca or crt)"
 
         if ca is not None:
+            if not os.path.exists(ca):
+                raise ValueError("The certificate file ({}) does not exist".format(ca))
+            self._assert_not_chain(ca)
+
             shutil.copyfile(ca, '%s/ca.crt' % self._certdb)
-            cmd = ['/usr/bin/c_rehash', self._certdb]
-            self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-            check_output(cmd, stderr=subprocess.STDOUT)
+            self.openssl_rehash(self._certdb)
             cmd = [
                 '/usr/bin/certutil',
                 '-A',
@@ -624,9 +813,15 @@ only.
                 '%s/%s' % (self._certdb, PWD_TXT),
             ]
             self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-            check_output(cmd, stderr=subprocess.STDOUT)
+            try:
+                check_output(cmd, stderr=subprocess.STDOUT)
+            except subprocess.CalledProcessError as e:
+                raise ValueError(e.output.decode('utf-8').rstrip())
 
         if crt is not None:
+            if not os.path.exists(crt):
+                raise ValueError("The certificate file ({}) does not exist".format(crt))
+            self._assert_not_chain(crt)
             cmd = [
                 '/usr/bin/certutil',
                 '-A',
@@ -639,7 +834,10 @@ only.
                 '%s/%s' % (self._certdb, PWD_TXT),
             ]
             self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-            check_output(cmd, stderr=subprocess.STDOUT)
+            try:
+                check_output(cmd, stderr=subprocess.STDOUT)
+            except subprocess.CalledProcessError as e:
+                raise ValueError(e.output.decode('utf-8').rstrip())
             cmd = [
                 '/usr/bin/certutil',
                 '-V',
@@ -648,7 +846,10 @@ only.
                 '-u', 'YCV'
             ]
             self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-            check_output(cmd, stderr=subprocess.STDOUT)
+            try:
+                check_output(cmd, stderr=subprocess.STDOUT)
+            except subprocess.CalledProcessError as e:
+                raise ValueError(e.output.decode('utf-8').rstrip())
 
     def create_rsa_user(self, name, months=VALID):
         """
@@ -691,8 +892,11 @@ only.
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
-        result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
         self.log.debug("nss output: %s", result)
         # Now extract this into PEM files that we can use.
         # pk12util -o user-william.p12 -d . -k pwdfile.txt -n user-william -W ''
@@ -705,7 +909,10 @@ only.
             '-W', '""'
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         # openssl pkcs12 -in user-william.p12 -passin pass:'' -out file.pem -nocerts -nodes
         # Extract the key
         cmd = [
@@ -718,7 +925,10 @@ only.
             '-nodes'
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         # Extract the cert
         cmd = [
             'openssl',
@@ -731,7 +941,10 @@ only.
             '-nodes'
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         # Convert the cert for userCertificate attr
         cmd = [
             'openssl',
@@ -742,7 +955,10 @@ only.
             '-out', '%s/%s%s.der' % (self._certdb, USER_PREFIX, name),
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
         return subject
 
@@ -756,6 +972,52 @@ only.
         crt_der_path = '%s/%s%s.der' % (self._certdb, USER_PREFIX, name)
         return {'ca': ca_path, 'key': key_path, 'crt': crt_path, 'crt_der_path': crt_der_path}
 
+    def list_keys(self, orphan=None):
+        key_list = []
+        cmd = [
+            '/usr/bin/certutil',
+            '-K',
+            '-d',
+            self._certdb,
+            '-f',
+            '%s/%s' % (self._certdb, PWD_TXT),
+        ]
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        # Ignore the first line of certutil output
+        for line in result.splitlines()[1:]:
+            # Normalise the output of certutil
+            line = re.sub(r"\<[^>]*\>","", line)
+            key = re.split(r'\s{2,}', line)
+            if orphan:
+                if 'orphan' in line:
+                    key_list.append(key)
+            else:
+                key_list.append(key)
+
+        return key_list
+
+    def del_key(self, keyid):
+        cmd = [
+            '/usr/bin/certutil',
+            '-F',
+            '-d',
+            self._certdb,
+            '-f',
+            '%s/%s' % (self._certdb, PWD_TXT),
+            '-k',
+            keyid,
+        ]
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        return result
+
     # Certificate helper functions
     def del_cert(self,  nickname):
         """Delete this certificate
@@ -769,7 +1031,10 @@ only.
                 '%s/%s' % (self._certdb, PWD_TXT),
             ]
         self.log.debug("del_cert cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
     def edit_cert_trust(self, nickname,  trust_flags):
         """Edit trust flags
@@ -799,7 +1064,10 @@ only.
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
         self.log.debug("edit_cert_trust cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
 
     def display_cert_details(self, nickname):
@@ -812,7 +1080,12 @@ only.
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
         self.log.debug("display_cert_details cmd: %s", format_cmd_list(cmd))
-        return check_output(cmd, stderr=subprocess.STDOUT, encoding='utf-8')
+        try:
+            result = check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        return ensure_str(result)
 
 
     def get_cert_details(self, nickname):
@@ -889,7 +1162,6 @@ only.
             if self._rsa_cert_is_caclienttrust(cert)
         ]
 
-
     def add_cert(self, nickname, input_file, ca=False):
         """Add server or CA cert
         """
@@ -898,9 +1170,21 @@ only.
         if not os.path.exists(input_file):
             raise ValueError("The certificate file ({}) does not exist".format(input_file))
 
+        pem_file = True
+        if not input_file.lower().endswith(".pem"):
+            pem_file = False
+        else:
+            self._assert_not_chain(input_file)
+
         if ca:
+            # Verify this is a CA cert
+            if not cert_is_ca(input_file):
+                raise ValueError(f"Certificate ({nickname}) is not a CA certificate")
             trust_flags = "CT,,"
         else:
+            # Verify this is a server cert
+            if cert_is_ca(input_file):
+                raise ValueError(f"Certificate ({nickname}) is not a server certificate")
             trust_flags = ",,"
 
         cmd = [
@@ -910,12 +1194,17 @@ only.
             '-n', nickname,
             '-t', trust_flags,
             '-i', input_file,
-            '-a',
             '-f',
             '%s/%s' % (self._certdb, PWD_TXT),
         ]
+        if pem_file:
+            cmd.append('-a')
+
         self.log.debug("add_cert cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
 
     def add_server_key_and_cert(self, input_key, input_cert):
         if not os.path.exists(input_key):
@@ -944,7 +1233,10 @@ only.
             '-aes128'
         ]
         self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-        check_output(cmd, stderr=subprocess.STDOUT)
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
         # Remove the server-cert if it exists, because else the import name fails.
         try:
             self.del_cert(CERT_NAME)
@@ -961,10 +1253,153 @@ only.
                 '-W', "",
             ]
             self.log.debug("nss cmd: %s", format_cmd_list(cmd))
-            check_output(cmd, stderr=subprocess.STDOUT)
+            try:
+                check_output(cmd, stderr=subprocess.STDOUT)
+            except subprocess.CalledProcessError as e:
+                raise ValueError(e.output.decode('utf-8').rstrip())
         finally:
             # Remove the p12
             if os.path.exists(p12_bundle):
                 os.remove(p12_bundle)
 
+    def add_ca_cert_bundle(self, cert_file, nicknames):
+        """
+        Add a PEM file that could be a bundle of CA certs
 
+        :param nicknames: list of names of each CA certificate
+        :param cert_file: path to certificate PEM file
+        :raises:
+        """
+
+        # Verify input_file exists
+        if not os.path.exists(cert_file):
+            raise ValueError("The certificate file ({}) does not exist".format(cert_file))
+
+        if not cert_file.lower().endswith(".pem"):
+            # Binary cert, this can not be a bundle
+            self.add_cert(nicknames[0], cert_file, ca=True)
+            log.info(f"Successfully added CA certificate ({nicknames[0]})")
+            return
+
+        try:
+            with open(cert_file, 'r') as f:
+                ca_count = 0
+                ca_files_to_cleanup = []
+                writing_ca = False
+                try:
+                    for line in f:
+                        if line.startswith('-----BEGIN CERTIFICATE-----'):
+                            # create tmp cert file
+                            writing_ca = True
+                            ca_file_name = cert_file + "-" + str(ca_count)
+                            ca_file = open(ca_file_name, 'w')
+                            ca_files_to_cleanup.append(ca_file_name)
+                            ca_file.write(line)
+                        elif writing_ca:
+                            ca_file.write(line)
+                            if line.startswith('-----END CERTIFICATE-----'):
+                                writing_ca = False
+                                ca_file.close()
+
+                                # Generate CA certificate nickname
+                                names_len = len(nicknames)
+                                if names_len > ca_count:
+                                    if nicknames[ca_count].lower() == CERT_NAME.lower() or nicknames[ca_count].lower() == CA_NAME.lower():
+                                        # cleanup
+                                        try:
+                                            for tmp_file in ca_files_to_cleanup:
+                                                os.remove(tmp_file)
+                                        except IOError as e:
+                                            # failed to remove tmp cert
+                                            log.debug("Failed to remove tmp cert file: " + str(e))
+                                        raise ValueError(f"You may not import a CA with the nickname {CERT_NAME} or {CA_NAME}")
+                                    ca_cert_name = nicknames[ca_count]
+                                else:
+                                    # A name was not provided for this cert, create one
+                                    ca_cert_name = nicknames[names_len - 1] + str(ca_count)
+
+                                # Check if certificate nickname exists
+                                name_exists = False
+                                try:
+                                    self.get_cert_details(ca_cert_name)
+                                    name_exists = True
+                                except ValueError:
+                                    pass
+
+                                if name_exists:
+                                    # Not good, cleanup and raise error
+                                    try:
+                                        for tmp_file in ca_files_to_cleanup:
+                                            os.remove(tmp_file)
+                                    except IOError as e:
+                                        # failed to remove tmp cert
+                                        log.debug("Failed to remove tmp cert file: " + str(e))
+                                    raise ValueError(f"Certificate already exists with the same name ({ca_cert_name})")
+
+                                # Verify this is a CA cert
+                                if not cert_is_ca(ca_file_name):
+                                    raise ValueError(f"Certificate ({ca_cert_name}) is not a CA certificate")
+
+                                # Add this CA certificate
+                                self.add_cert(ca_cert_name, ca_file_name, ca=True)
+                                ca_count += 1
+                                log.info(f"Successfully added CA certificate ({ca_cert_name})")
+
+                    # All done, cleanup
+                    try:
+                        for tmp_file in ca_files_to_cleanup:
+                            os.remove(tmp_file)
+                    except IOError as e:
+                        # failed to remove tmp cert
+                        log.debug("Failed to remove tmp cert file: " + str(e))
+                except IOError as e:
+                    # Some failure, remove all the tmp files
+                    ca_file.close()
+                    try:
+                        for tmp_file in ca_files_to_cleanup:
+                            os.remove(tmp_file)
+                    except IOError as e:
+                        # failed to remove tmp cert
+                        log.debug("Failed to remove tmp cert file: " + str(e))
+                    raise e
+        except EnvironmentError as e:
+            raise e
+
+    def export_cert(self, nickname, output_file=None, der_format=False):
+        """
+        :param nickname: name of certificate
+        :param output_file: name for exported certificate
+        :param der_format: export certificate in DER/binary format
+        :raise ValueError: error
+        """
+        cmd = [
+            '/usr/bin/certutil',
+            '-L',
+            '-d', self._certdb,
+            '-n', nickname,
+            '-f',
+            '%s/%s' % (self._certdb, PWD_TXT),
+        ]
+
+        # Handle args for PEM vs DER options
+        if der_format:
+            cmd.append('-r')
+            if output_file is None:
+                output_file = nickname + ".crt"
+        else:
+            cmd.append('-a')
+            if output_file is None:
+                output_file = nickname + ".pem"
+
+        # Set output file name
+        if not output_file.startswith("/"):
+            # Must be full path, otherwise but it in the cert dir
+            output_file = f"{self._certdb}/{output_file}"
+        cmd.append('-o')
+        cmd.append(output_file)
+
+        self.log.debug("export_cert cmd: %s", format_cmd_list(cmd))
+        try:
+            check_output(cmd, stderr=subprocess.STDOUT)
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
