@@ -45,7 +45,8 @@ import errno
 from socket import getfqdn
 from ldapurl import LDAPUrl
 from contextlib import closing
-
+from cryptography import x509
+from cryptography.hazmat.backends import default_backend
 import lib389
 from pathlib import Path
 from lib389.paths import ( Paths, DEFAULTS_PATH )
@@ -495,7 +496,10 @@ def is_dn_parent(parent_dn, child_dn, ):
 def normalizeDN(dn, usespace=False):
     # not great, but will do until we use a newer version of python-ldap
     # that has DN utilities
-    ary = ldap.explode_dn(dn.lower())
+    try:
+        ary = ldap.explode_dn(dn.lower())
+    except ldap.DECODING_ERROR:
+        raise ValueError(f"Unable to normalize DN '{dn}'")
     joinstr = ","
     if usespace:
         joinstr = ", "
@@ -1306,7 +1310,7 @@ def ds_is_related(relation, *ver, instance=None):
     if len(ver) > 1:
         for cmp_ver in ver:
             if cmp_ver.startswith(ds_ver[:3]):
-                return ops[relation](DSVersion(ds_ver),DSVersion(cmp_ver))
+                return ops[relation](DSVersion(ds_ver), DSVersion(cmp_ver))
     else:
         return ops[relation](DSVersion(ds_ver), DSVersion(ver[0]))
 
@@ -1831,3 +1835,43 @@ def get_task_status(inst, log, taskObj, dn=None, show_log=False, watch=False, us
         elif not use_json:
             log.info("Refreshing status ...")
         time.sleep(2)
+
+
+def is_cert_der(file_name):
+    try:
+        with open(file_name, 'tr') as check_file:
+            check_file.read()
+            # PEM cert
+            return False
+    except UnicodeDecodeError:
+        # DER cert
+        return True
+
+
+def cert_is_ca(cert_file_name):
+    with open(cert_file_name, "rb") as f:
+        if is_cert_der(cert_file_name):
+            cert = x509.load_der_x509_certificate(f.read(), default_backend())
+        else:
+            cert = x509.load_pem_x509_certificate(f.read(), default_backend())
+
+    key_usage = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.KEY_USAGE)
+    if not key_usage.value.key_cert_sign:
+        return False
+
+    basic_constraints = cert.extensions.get_extension_for_oid(
+        x509.oid.ExtensionOID.BASIC_CONSTRAINTS
+    )
+    if not basic_constraints.value.ca:
+        return False
+
+    # This is a CA cert
+    return True
+
+
+def get_passwd_from_file(passwd_file):
+    if os.path.exists(passwd_file):
+        with open(passwd_file, 'r') as f:
+            passwd = f.readline().strip()
+            return passwd
+    raise ValueError(f"The password file '{passwd_file}' does not exist, or can not be read.")
